// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Zone Memory Allocation. Neat.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: z_zone.c,v 1.4 1997/02/03 16:47:58 b1 Exp $";

#include "z_zone.h"
#include "i_system.h"
#include "doomdef.h"


#include "pnobridge.h"
// idx is lump, 0 is pointer, 1 is tag, 2 is user data pointer
extern UInt32 gblFeatureArray[3 * FEATURE_ARRAY_COUNT];

//
// ZONE MEMORY ALLOCATION
//
// There is never any space between memblocks,
//  and there will never be two contiguous ydk_free memblocks.
// The rover can be left pointing at a non-empty block.
//
// It is of no value to ydk_free a cachable block,
//  because it will get overwritten automatically if needed.
// 
 
#define ZONEID	0x1d4a11


typedef struct
{
    // total bytes malloced, including header
    int		size;

    // start / end cap for linked list
    memblock_t	blocklist;
    
    memblock_t*	rover;
    
} memzone_t;



memzone_t*	mainzone;



//
// Z_ClearZone
//
void Z_ClearZone (memzone_t* zone)
{
    memblock_t*		block;
	
    // set the entire zone to one ydk_free block
    zone->blocklist.next =
	zone->blocklist.prev =
	block = (memblock_t *)( (byte *)zone + sizeof(memzone_t) );
    
    zone->blocklist.user = (void *)zone;
    zone->blocklist.tag = PU_STATIC;
    zone->rover = block;
	
    block->prev = block->next = &zone->blocklist;
    
    // NULL indicates a ydk_free block.
    block->user = NULL;	

    block->size = zone->size - sizeof(memzone_t);
}



//
// Z_Init
//
void Z_Init (void)
{
    memblock_t*	block;
    int		size;

    mainzone = (memzone_t *)I_ZoneBase (&size);
    mainzone->size = size;

    // set the entire zone to one ydk_free block
    mainzone->blocklist.next =
	mainzone->blocklist.prev =
	block = (memblock_t *)( (byte *)mainzone + sizeof(memzone_t) );

    mainzone->blocklist.user = (void *)mainzone;
    mainzone->blocklist.tag = PU_STATIC;
    mainzone->rover = block;
	
    block->prev = block->next = &mainzone->blocklist;

    // NULL indicates a ydk_free block.
    block->user = NULL;
    
    block->size = mainzone->size - sizeof(memzone_t);
}


//
// Z_Free
//
void Z_Free (void* ptr)
{
    memblock_t*		block;

    block = (memblock_t *) ( (byte *)ptr - sizeof(memblock_t));

    if (block->id != ZONEID)
    {
		// first check feature array
		UInt32 lump = ((UInt32 *) ptr)[-1];
		if ((lump > 0) && (lump < FEATURE_ARRAY_COUNT))
		{
			int k = lump * 3;
			if (gblFeatureArray[k + 0] == (UInt32) ptr)
			{
				if (gblFeatureArray[k + 2] > 0x100)
				{
					void **user = (void **) gblFeatureArray[k + 2];
					*user = NULL;
				}
				gblFeatureArray[k + 0] = 0;
				gblFeatureArray[k + 1] = 0;
				gblFeatureArray[k + 2] = 0;
				FtrPtrFree(CREATORID, FTR_NUM_LUMPBASE + lump);
				return;
			}
		}
		I_Error ("Z_Free: freed a pointer without ZONEID");
		return;
	}
	
	Z_Free2(ptr, true); // update user pointer
}

void Z_Free2(void *ptr, int bUpdateUserPtr)
{
    memblock_t*		block;
    memblock_t*		other;

    block = (memblock_t *) ( (byte *)ptr - sizeof(memblock_t));
    if (block->id != ZONEID)
    	return;

	if (bUpdateUserPtr)
	{
	    if (block->user > (void **)0x100)
	    {
			// smaller values are not pointers
			// Note: OS-dependend?
			// clear the user's mark
			*block->user = 0;
		}
	}

    // mark as ydk_free
    block->user = NULL;	
    block->tag = 0;
    block->id = 0;
	
    other = block->prev;

    if (!other->user)
    {
	// merge with previous ydk_free block
	other->size += block->size;
	other->next = block->next;
	other->next->prev = other;

	if (block == mainzone->rover)
	    mainzone->rover = other;

	block = other;
    }
	
    other = block->next;
    if (!other->user)
    {
	// merge the next ydk_free block onto the end
	block->size += other->size;
	block->next = other->next;
	block->next->prev = block;

	if (other == mainzone->rover)
	    mainzone->rover = block;
    }
}



//
// Z_Malloc
// You can pass a NULL user if the tag is < PU_PURGELEVEL.
//
#define MINFRAGMENT		64


void*
Z_Malloc
( int		size,
  int		tag,
  void*		user )
{
    int		extra;
    memblock_t*	start;
    memblock_t* rover;
    memblock_t* newblock;
    memblock_t*	base;

    size = (size + 3) & ~3;
    
    // scan through the block list,
    // looking for the first ydk_free block
    // of sufficient size,
    // throwing out any purgable blocks along the way.

    // account for size of block header
    size += sizeof(memblock_t);
    
    // if there is a ydk_free block behind the rover,
    //  back up over them
    base = mainzone->rover;
    
    if (!base->prev->user)
	base = base->prev;
	
    rover = base;
    start = base->prev;
	
    do
    {
	if (rover == start)
	{
	    // scanned all the way around the list
	    I_Error ("Z_Malloc: failed on allocation of %i bytes", size);
	}
	
	if (rover->user)
	{
	    if (rover->tag < PU_PURGELEVEL)
	    {
		// hit a block that can't be purged,
		//  so move base past it
		base = rover = rover->next;
	    }
	    else
	    {
		// ydk_free the rover block (adding the size to base)

		// the rover can be the base block
		base = base->prev;
		Z_Free ((byte *)rover+sizeof(memblock_t));
		base = base->next;
		rover = base->next;
	    }
	}
	else
	    rover = rover->next;
    } while (base->user || base->size < size);

    
    // found a block big enough
    extra = base->size - size;
    
    if (extra >  MINFRAGMENT)
    {
	// there will be a ydk_free fragment after the allocated block
	newblock = (memblock_t *) ((byte *)base + size );
	newblock->size = extra;
	
	// NULL indicates ydk_free block.
	newblock->user = NULL;	
	newblock->tag = 0;
	newblock->prev = base;
	newblock->next = base->next;
	newblock->next->prev = newblock;

	base->next = newblock;
	base->size = size;
    }
	
    if (user)
    {
	// mark as an in use block
	base->user = user;			
	*(void **)user = (void *) ((byte *)base + sizeof(memblock_t));
    }
    else
    {
	if (tag >= PU_PURGELEVEL)
	    I_Error ("Z_Malloc: an owner is required for purgable blocks");

	// mark as in use, but unowned	
	base->user = (void *)2;		
    }
    base->tag = tag;

    // next allocation will start looking here
    mainzone->rover = base->next;	
	
    base->id = ZONEID;
    
    return (void *) ((byte *)base + sizeof(memblock_t));
}



//
// Z_FreeTags
//
void
Z_FreeTags
( int		lowtag,
  int		hightag )
{
    memblock_t*	block;
    memblock_t*	next;
    
	// first check the storage space...
	int i, k;
	k = 0;
	for (i=0; i<FEATURE_ARRAY_COUNT; i++)
	{
		int t = gblFeatureArray[k + 1];
		if (t >= lowtag && t <= hightag)
		{
			if (gblFeatureArray[k + 2] > 0x100)
			{
				void **user = (void **) gblFeatureArray[k + 2];
				*user = NULL;
			}
			gblFeatureArray[k + 0] = 0;
			gblFeatureArray[k + 1] = 0;
			gblFeatureArray[k + 2] = 0;
			FtrPtrFree(CREATORID, FTR_NUM_LUMPBASE + i);
		}
		k += 3;
	}
  
    for (block = mainzone->blocklist.next ;
	 block != &mainzone->blocklist ;
	 block = next)
    {
	// get link before freeing

	next = block->next;


	// ydk_free block?
	if (!block->user)
	    continue;

	if (block->tag >= lowtag && block->tag <= hightag)
	{
	    Z_Free ( (byte *)block+sizeof(memblock_t));
	   }
    }
}



//
// Z_DumpHeap
// Note: TFileDumpHeap( stdout ) ?
//
void
Z_DumpHeap
( int		lowtag,
  int		hightag )
{
    memblock_t*	block;
	

    for (block = mainzone->blocklist.next ; ; block = block->next)
    {
	if (block->tag >= lowtag && block->tag <= hightag){
	    Sys_PrintF ("b:%p s:%7i u:%p tag:%3i\n",
		    block, block->size, block->user, block->tag);
		}
	if (block->next == &mainzone->blocklist)
	{
	    // all blocks have been hit
	    break;
	}
	
	if ( (byte *)block + block->size != (byte *)block->next){
	    Sys_PrintF ("ERROR: block size does not touch the next block\n");
	    menu_info(" ");
	    Sys_PrintF ("b:%p s:%7i u:%p tag:%3i\n",
		    block, block->size, block->user, block->tag);
		}

	if ( block->next->prev != block){
	    Sys_PrintF ("ERROR: next block doesn't have proper back link\n");
	    menu_info(" ");
	    Sys_PrintF ("b:%p s:%7i u:%p tag:%3i\n",
		    block, block->size, block->user, block->tag);
		menu_info(" ");}

	if (!block->user && !block->next->user){
	    Sys_PrintF ("ERROR: two consecutive free blocks\n");
	    }
    }
}


//
// Z_FileDumpHeap
//
void Z_FileDumpHeap (FILE* f)
{
    memblock_t*	block;
	
    fprintf (f,"zone size: %i  location: %p\n",mainzone->size,mainzone);
	
    for (block = mainzone->blocklist.next ; ; block = block->next)
    {
	fprintf (f,"block:%p    size:%7i    user:%p    tag:%3i\n",
		 block, block->size, block->user, block->tag);
		
	if (block->next == &mainzone->blocklist)
	{
	    // all blocks have been hit
	    break;
	}
	
	if ( (byte *)block + block->size != (byte *)block->next)
	    fprintf (f,"ERROR: block size does not touch the next block\n");

	if ( block->next->prev != block)
	    fprintf (f,"ERROR: next block doesn't have proper back link\n");

	if (!block->user && !block->next->user)
	    fprintf (f,"ERROR: two consecutive ydk_free blocks\n");
    }
}



//
// Z_CheckHeap
//
void Z_CheckHeap (void)
{
    memblock_t*	block;
	
    for (block = mainzone->blocklist.next ; ; block = block->next)
    {
	if (block->next == &mainzone->blocklist)
	{
	    // all blocks have been hit
	    break;
	}
	
	if ( (byte *)block + block->size != (byte *)block->next)
	    I_Error ("Z_CheckHeap: block size does not touch the next block\n");

	if ( block->next->prev != block)
	    I_Error ("Z_CheckHeap: next block doesn't have proper back link\n");

	if (!block->user && !block->next->user)
	    I_Error ("Z_CheckHeap: two consecutive ydk_free blocks\n");
    }
}


//
// Z_ChangeTag
//
static void Z_ChangeTag3(void* ptr, int tag)
{
    memblock_t*	block;
	
    block = (memblock_t *) ( (byte *)ptr - sizeof(memblock_t));

    if (block->id != ZONEID)
    {
	I_Error ("Z_ChangeTag: freed a pointer without ZONEID");
	return;
	}

    if (tag >= PU_PURGELEVEL && (unsigned)block->user < 0x100)
    {
	I_Error ("Z_ChangeTag: an owner is required for purgable blocks");
	return;
	}

    block->tag = tag;
}


void Z_ChangeTag2(void *p, int t, char *file, int line)
{
	if (p)
	{
		if (( (memblock_t *)( (byte *)(p) - sizeof(memblock_t)))->id != ZONEID)
		{
			// first check feature array
			UInt32 lump = ((UInt32 *) p)[-1];
			if ((lump > 0) && (lump < FEATURE_ARRAY_COUNT))
			{
				int k = lump * 3;
				if (gblFeatureArray[k + 0] == (UInt32) p)
				{
					gblFeatureArray[k + 1] = t;
//					return;
				}
				else
				{
					Sys_PrintF("ChangeTag %lx %lx  %s:%ld", gblFeatureArray[k + 0], p, file, line);
					menu_info("ok");
				}
			}
			// whoe knows where this comes from... Orphan pointers?
//			Sys_PrintF("Z_CT (%lx) at %s : %d", p, file, line);
//			menu_info("ok");
		}
		else
		{
			Z_ChangeTag3(p,t);
		}
	}
}



//
// Z_FreeMemory
//
int Z_FreeMemory (void)
{
    memblock_t*		block;
    int			ydk_free;
	
    ydk_free = 0;
    
    for (block = mainzone->blocklist.next ;
	 block != &mainzone->blocklist;
	 block = block->next)
    {
	if (!block->user || block->tag >= PU_PURGELEVEL)
	    ydk_free += block->size;
    }
    return ydk_free;
}

